# Copyright 2020, Charles Powell

import hashlib
from time import time
from typing import Optional, Any
from functools import lru_cache


@lru_cache(maxsize=256)
def _generate_mac(device_id: str) -> str:
    """Generate a MAC address from the device ID only once."""
    end = [int(device_id[i : i + 2], 16) for i in range(0, len(device_id), 2)]
    mac = [53, 75, 31] + end[:3]
    return ":".join("%02X" % b for b in mac)


@lru_cache(maxsize=256)
def _generate_device_id(id: str) -> str:
    """Generate a device ID from the ID only once."""
    return hashlib.sha1(id.encode("utf-8")).hexdigest().upper()


class PlugInstance:
    """Class to represent a single plug instance."""

    __slots__ = (
        "id",
        "voltage",
        "power",
        "current",
        "alias",
        "start_time",
        "device_id",
        "mac",
    )

    def __init__(
        self,
        id: str,
        start_time: Optional[float] = None,
        alias: Optional[str] = None,
        power=0.0,
        current=0.0,
        voltage=120.0,
        mac: Optional[str] = None,
        device_id: Optional[str] = None,
    ) -> None:
        """Initialize a plug instance."""
        self.id = id
        self.voltage = voltage
        self.power = power
        self.current = current
        if not self.power:
            self.power = self.voltage * self.current
        if not self.current:
            self.current = self.power / self.voltage
        self.alias = alias or id
        self.start_time = start_time or time() - 1
        if device_id:
            self.device_id = device_id.upper()
        else:
            self.device_id = _generate_device_id(id)
        if mac:
            self.mac = mac.upper()
        else:
            self.mac = _generate_mac(self.device_id)

    def generate_response(self) -> dict[str, dict[str, Any]]:
        """Generate a response dict for the plug."""
        # Response dict
        return {
            "emeter": {
                "get_realtime": {
                    "current": self.current,
                    "voltage": self.voltage,
                    "power": self.power,
                    "total": 0,  # Unsure if this needs a value, appears not to
                    "err_code": 0,  # No errors here!
                }
            },
            "system": {
                "get_sysinfo": {
                    "err_code": 0,
                    "sw_ver": "1.2.5 Build 171206 Rel.085954",
                    "hw_ver": "1.0",
                    "type": "IOT.SMARTPLUGSWITCH",
                    "model": "HS110(US)",
                    "mac": self.mac,
                    "deviceId": self.device_id,
                    "hwId": "60FF6B258734EA6880E186F8C96DDC61",
                    "fwId": "00000000000000000000000000000000",
                    "oemId": "FFF22CFF774A0B89F7624BFC6F50D5DE",
                    "alias": self.alias,
                    "dev_name": "Wi-Fi Smart Plug With Energy Monitoring",
                    "icon_hash": "",
                    "relay_state": 1 if self.power > 0 else 0,
                    "on_time": time() - self.start_time,
                    "active_mode": "none",
                    "feature": "TIM:ENE",
                    "updating": 0,
                    "rssi": -60,  # Great wifi signal
                    "led_off": 0,  # Probably not important
                    "latitude": 39.8283,  # Center of the US
                    "longitude": -98.5795,  # Center of the US
                }
            },
        }



class SenseAPITimeoutException(Exception):
    pass


class SenseAuthenticationException(Exception):
    pass


class SenseMFARequiredException(Exception):
    pass


class SenseWebsocketException(Exception):
    pass


class SenseAPIException(Exception):
    pass



import asyncio
import ssl
import sys
from functools import lru_cache
from time import time
from datetime import timezone

import aiohttp
import orjson
import websockets

from .sense_api import *
from .sense_exceptions import *

if sys.version_info[:2] < (3, 11):
    from async_timeout import timeout as asyncio_timeout
else:
    from asyncio import timeout as asyncio_timeout


@lru_cache(maxsize=None)
def get_ssl_context(ssl_verify: bool, ssl_cafile: str) -> ssl.SSLContext:
    """Create or set the SSL context. Use custom ssl verification, if specified."""
    if not ssl_verify:
        ssl_context = ssl.create_default_context()
        ssl_context.check_hostname = False
        ssl_context.verify_mode = ssl.CERT_NONE
    elif ssl_cafile:
        ssl_context = ssl.create_default_context(cafile=ssl_cafile)
    else:
        ssl_context = ssl.create_default_context()
    return ssl_context


class ASyncSenseable(SenseableBase):
    def __init__(
        self,
        username=None,
        password=None,
        api_timeout=API_TIMEOUT,
        wss_timeout=WSS_TIMEOUT,
        client_session=None,
        ssl_verify=True,
        ssl_cafile="",
        device_id=None,
    ):
        """Init the ASyncSenseable object."""
        self._client_session = client_session or aiohttp.ClientSession()

        super().__init__(
            username=username,
            password=password,
            api_timeout=api_timeout,
            wss_timeout=wss_timeout,
            ssl_verify=ssl_verify,
            ssl_cafile=ssl_cafile,
            device_id=device_id,
        )

    def set_ssl_context(self, ssl_verify: bool, ssl_cafile: str):
        """Create or set the SSL context. Use custom ssl verification, if specified."""
        self.ssl_context = get_ssl_context(ssl_verify, ssl_cafile)

    async def authenticate(self, username: str, password: str, ssl_verify: bool = True, ssl_cafile: str = "") -> None:
        """Authenticate with username (email) and password. Optionally set SSL context as well.
        This or `load_auth` must be called once at the start of the session."""
        auth_data = {"email": username, "password": password}
        self.set_ssl_context(ssl_verify, ssl_cafile)

        # Get auth token
        async with self._client_session.post(
            API_URL + "authenticate",
            headers=self.headers,
            timeout=self.api_timeout,
            data=auth_data,
        ) as resp:
            # check MFA code required
            if resp.status == 401:
                data = await resp.json()
                if "mfa_token" in data:
                    self._mfa_token = data["mfa_token"]
                    raise SenseMFARequiredException(data["error_reason"])

            # check for 200 return
            if resp.status != 200:
                raise SenseAuthenticationException(
                    f"Please check username and password. API Return Code: {resp.status}"
                )

            # Build out some common variables
            data = await resp.json()
            self._set_auth_data(data)
            self.set_monitor_id(data["monitors"][0]["id"])
            await self.fetch_devices()

    async def validate_mfa(self, code: str) -> None:
        """Validate a multi-factor authentication code after authenticate raised SenseMFARequiredException.
        Authentication process is completed if code is valid."""
        mfa_data = {
            "totp": code,
            "mfa_token": self._mfa_token,
            "client_time:": datetime.now(timezone.utc).strftime("%Y-%m-%dT%H:%M:%SZ"),
        }

        # Get auth token
        async with self._client_session.post(
            API_URL + "authenticate/mfa",
            headers=self.headers,
            timeout=self.api_timeout,
            data=mfa_data,
        ) as resp:
            # check for 200 return
            if resp.status != 200:
                raise SenseAuthenticationException(f"API Return Code: {resp.status}")

            # Build out some common variables
            data = await resp.json()
            self._set_auth_data(data)
            self.set_monitor_id(data["monitors"][0]["id"])
            await self.fetch_discovered_devices()

    async def renew_auth(self) -> None:
        """Renew the authentication token."""
        renew_data = {
            "user_id": self.sense_user_id,
            "refresh_token": self.refresh_token,
        }

        # Get auth token
        async with self._client_session.post(
            API_URL + "renew",
            headers=self.headers,
            timeout=self.api_timeout,
            data=renew_data,
        ) as resp:
            # check for 200 return
            if resp.status != 200:
                raise SenseAuthenticationException(f"API Return Code: {resp.status}")

            self._set_auth_data(await resp.json())

    async def logout(self) -> None:
        """Log out of Sense."""
        # Get auth token
        async with self._client_session.get(API_URL + "logout", timeout=self.api_timeout) as resp:
            # check for 200 return
            if resp.status != 200:
                raise SenseAPIException(f"API Return Code: {resp.status}")

    async def update_realtime(self, retry: bool = True) -> None:
        """Update the realtime data (device status and current power)."""
        # rate limit API calls
        now = time()
        if self._realtime and self.rate_limit and self.last_realtime_call + self.rate_limit > now:
            return self._realtime
        self.last_realtime_call = now
        try:
            await self.async_realtime_stream(single=True)
        except SenseAuthenticationException as e:
            if retry:
                await self.renew_auth()
                await self.update_realtime(False)
            else:
                raise e

    async def async_realtime_stream(self, callback: callable = None, single: bool = False) -> None:
        """Reads realtime data from websocket.  Data is passed to callback if available.
        Continues reading realtime stream data forever unless 'single' is set to True.
        """
        url = WS_URL % (self.sense_monitor_id, self.sense_access_token)
        # hello, features, [updates,] data
        async with websockets.connect(url, ssl=self.ssl_context) as ws:
            while True:
                try:
                    async with asyncio_timeout(self.wss_timeout):
                        message = await ws.recv()
                except asyncio.TimeoutError as ex:
                    raise SenseAPITimeoutException("API websocket timed out") from ex

                result = orjson.loads(message)
                if result.get("type") == "realtime_update":
                    data = result["payload"]
                    self._set_realtime(data)
                    if callback:
                        callback(data)
                    if single:
                        return
                elif result.get("type") == "error":
                    data = result["payload"]
                    if not data["authorized"]:
                        raise SenseAuthenticationException("Web Socket Unauthorized")
                    raise SenseWebsocketException(data["error_reason"])

    async def get_realtime_future(self, callback: callable) -> None:
        """Returns an async Future to parse realtime data with callback"""
        await self.async_realtime_stream(callback)

    async def _api_call(self, url, payload={}, retry=False):
        """Make a call to the Sense API directly and return the json results."""
        timeout = aiohttp.ClientTimeout(total=self.api_timeout)
        try:
            async with self._client_session.get(
                API_URL + url, headers=self.headers, timeout=timeout, data=payload
            ) as resp:
                if not retry and resp.status == 401:
                    await self.renew_auth()
                    return await self._api_call(url, payload, True)

                # 4xx represents unauthenticated
                if resp.status == 401 or resp.status == 403 or resp.status == 404:
                    raise SenseAuthenticationException(f"API Return Code: {resp.status}")

                if resp.status != 200:
                    raise SenseAPIException(f"API Return Code: {resp.status}")

                return await resp.json()
        except asyncio.TimeoutError as ex:
            # timed out
            raise SenseAPITimeoutException("API call timed out") from ex

    async def get_trend_data(self, scale: Scale, dt: datetime = None) -> None:
        """Update trend data for specified scale from API.
        Optionally set a date to fetch data from."""
        if not dt:
            dt = datetime.now(timezone.utc)
        json = self._api_call(
            f"app/history/trends?monitor_id={self.sense_monitor_id}"
            + f"&device_id=always_on&scale={scale.name}&start={dt.strftime('%Y-%m-%dT%H:%M:%S')}"
        )
        self._trend_data[scale] = await json
        self._update_device_trends(scale)

    async def update_trend_data(self, dt: datetime = None) -> None:
        """Update trend data of all scales from API.
        Optionally set a date to fetch data from."""
        for scale in Scale:
            await self.get_trend_data(scale, dt)

    async def get_monitor_data(self):
        """Get monitor overview info from API."""
        json = await self._api_call(f"app/monitors/{self.sense_monitor_id}/overview")
        if "monitor_overview" in json and "monitor" in json["monitor_overview"]:
            self._monitor = json["monitor_overview"]["monitor"]
        return self._monitor

    async def fetch_devices(self) -> None:
        """Fetch discovered devices from API."""
        json = await self._api_call(f"app/monitors/{self.sense_monitor_id}/devices/overview")
        smart_plugs = {}
        for device in json["devices"]:
            if not device["tags"].get("DeviceListAllowed", True) or device["tags"].get("MergeId", None):
                continue

            # handle smart plugs after
            if device["tags"].get("DefaultUserDeviceType", "") == "SmartPlug":
                if device["name"] not in smart_plugs:
                    smart_plugs[device["name"]] = []
                smart_plugs[device["name"]].append(device)
                continue

            id = device["id"]
            if id not in self._devices:
                self._devices[id] = SenseDevice(id)
            self._devices[id].name = device["name"]
            self._devices[id].icon = device["icon"]

        for devices in smart_plugs.values():
            device = {"tags": {"DateCreated": "2010-01-01T01:01:01.000Z"}}
            for d in devices:
                if d["tags"].get("DateCreated", "") > device["tags"]["DateCreated"]:
                    device = d

            id = device["id"]
            if id not in self._devices:
                self._devices[id] = SenseDevice(id)
            self._devices[id].name = device["name"]
            self._devices[id].icon = device["icon"]

    async def get_discovered_device_names(self) -> list[str]:
        """Outdated. Get list of device names from API.
        Use fetch_discovered_devices and sense.devices instead."""
        await self.fetch_devices()
        return [d.name for d in self._devices.values()]

    async def get_discovered_device_data(self):
        """Outdated. Get list of raw device data from API.
        Use fetch_discovered_devices and sense.devices instead."""
        json = self._api_call(f"monitors/{self.sense_monitor_id}/devices/overview")
        return await json["devices"]



# Copyright 2020, Charles Powell

# SenseLink is a tool that emulates the energy monitoring functionality of TP-Link Kasa HS110 Smart Plugs,
# and allows you to report "custom" power usage to your Sense Home Energy Monitor based on other parameters.

import asyncio
import logging
from typing import Optional, Union

import orjson
from kasa_crypt import decrypt as tp_link_decrypt
from kasa_crypt import encrypt as tp_link_encrypt

from .plug_instance import PlugInstance

SENSE_TP_LINK_PORT = 9999

_LOGGER = logging.getLogger(__name__)


class SenseLinkServerProtocol:
    """Class to represent a SenseLink server."""

    def __init__(self, devices: callable) -> None:
        """Initialize the SenseLink server."""
        self._devices = devices
        self.should_respond = True
        self.transport: Optional[asyncio.DatagramTransport] = None

    def connection_made(self, transport: asyncio.DatagramTransport) -> None:
        """Handle new connection."""
        self.transport = transport

    def connection_lost(self, exc) -> None:
        """Handle lost connection."""
        pass

    def datagram_received(self, data: bytes, addr: Union[tuple[str, int], tuple[str, int, int, int]]) -> None:
        """Handle incoming UDP datagram."""
        try:
            decrypted_data = tp_link_decrypt(data)
        except UnicodeDecodeError:
            _LOGGER.debug(f"Failed to decrypt data from {addr}")
            return

        try:
            json_data = orjson.loads(decrypted_data)
            # Sense requests the emeter and system parameters
            if (
                "emeter" in json_data
                and "get_realtime" in json_data["emeter"]
                and "system" in json_data
                and "get_sysinfo" in json_data["system"]
            ):
                # Check for non-empty values, to prevent echo storms
                if json_data["emeter"]["get_realtime"]:
                    # This is a self-echo, common with Docker without --net=Host!
                    logging.debug("Ignoring non-empty/non-Sense UDP request")
                    return

                logging.debug(f"Broadcast received from {addr}: {json_data}")

                # Build and send responses
                for plug in self._devices():
                    # Build response
                    response = plug.generate_response()
                    json_resp = orjson.dumps(response)
                    encrypted_resp = tp_link_encrypt(json_resp.decode("utf-8"))
                    # Strip leading 4 bytes for...some reason
                    encrypted_resp = encrypted_resp[4:]

                    # Allow disabling response
                    if self.should_respond:
                        # Send response
                        logging.debug("Sending response: %s", response)
                        self.transport.sendto(encrypted_resp, addr)
                    else:
                        # Do not send response, but log for debugging
                        _LOGGER.debug("SENSE_RESPONSE disabled, response content: %s", response)
            else:
                _LOGGER.debug(f"Ignoring non-emeter JSON from %s: %s", addr, json_data)

        # Appears to not be JSON
        except ValueError:
            _LOGGER.debug("Did not receive valid json")


class SenseLink:
    """Class to represent a SenseLink server."""

    _devices = []

    def __init__(self, devices: callable, port=SENSE_TP_LINK_PORT) -> None:
        """Initialize the SenseLink server."""
        self.port = port
        self._devices = devices

    def print_instance_wattages(self) -> None:
        """Log the current wattages of all instances."""
        for inst in self._devices():
            logging.info(f"Plug {inst.alias} power: {inst.power}")

    async def start(self) -> None:
        """Start the SenseLink server."""
        loop = asyncio.get_running_loop()
        self.transport, self.protocol = await loop.create_datagram_endpoint(
            lambda: SenseLinkServerProtocol(self._devices), local_addr=("0.0.0.0", self.port)
        )

    async def stop(self) -> None:
        """Stop the SenseLink server."""
        self.transport.close()



from .sense_api import SenseableBase, Scale, SenseDevice
from .sense_exceptions import *

from .senseable import Senseable
from .asyncsenseable import ASyncSenseable
from .plug_instance import PlugInstance
from .sense_link import SenseLink

__version__ = "0.13.1"



from enum import Enum, auto
from datetime import datetime
from typing import Optional
import ciso8601
import uuid
from .sense_exceptions import *

API_URL = "https://api.sense.com/apiservice/api/v1/"
WS_URL = "wss://clientrt.sense.com/monitors/%s/realtimefeed?access_token=%s"
API_TIMEOUT = 5
WSS_TIMEOUT = 5
RATE_LIMIT = 60


class Scale(Enum):
    DAY = auto()
    WEEK = auto()
    MONTH = auto()
    YEAR = auto()
    CYCLE = auto()


class SenseDevice:
    def __init__(self, id):
        self.id = id
        self.name = ""
        self.icon = ""
        self.is_on = False
        self.power_w = 0.0
        self.energy_kwh = {}
        for scale in Scale:
            self.energy_kwh[scale] = 0.0


class SenseableBase(object):
    def __init__(
        self,
        username: str = None,
        password: str = None,
        api_timeout: int = API_TIMEOUT,
        wss_timeout: int = WSS_TIMEOUT,
        ssl_verify: bool = True,
        ssl_cafile: str = "",
        device_id: str = None,
    ):
        """Initialize SenseableBase object."""

        # Timeout instance variables
        self.api_timeout = api_timeout
        self.wss_timeout = wss_timeout
        self.rate_limit = RATE_LIMIT
        self.last_realtime_call = 0

        self._mfa_token = ""
        self._realtime = {}
        self._devices: dict[str, SenseDevice] = {}
        self._trend_data: dict[Scale, dict] = {}
        self._monitor = {}
        for scale in Scale:
            self._trend_data[scale] = {}
        self.set_ssl_context(ssl_verify, ssl_cafile)
        if device_id:
            self.device_id = device_id
        else:
            self.device_id = str(uuid.uuid4()).replace("-", "")

        self.headers = {"x-sense-device-id": self.device_id}

        if username and password:
            self.authenticate(username, password)

    def load_auth(self, access_token: str, user_id: str, device_id: str, refresh_token: str):
        """Load the authentication data from a previous session."""
        self.device_id = device_id
        data = {
            "access_token": access_token,
            "user_id": user_id,
            "refresh_token": refresh_token,
        }
        self._set_auth_data(data)

    def set_monitor_id(self, monitor_id: str):
        self.sense_monitor_id = monitor_id

    def _set_auth_data(self, data):
        """Set the authentication data for the session."""
        self.sense_access_token = data["access_token"]
        self.sense_user_id = data["user_id"]
        self.refresh_token = data["refresh_token"]

        # create the auth header
        self.headers = {
            "x-sense-device-id": self.device_id,
            "Authorization": "bearer {}".format(self.sense_access_token),
        }

    def _update_device_trends(self, scale: Scale):
        if not self._trend_data[scale]["consumption"].get("devices"):
            return
        for d in self._devices.values():
            d.energy_kwh[scale] = 0
        for d in self._trend_data[scale]["consumption"]["devices"]:
            id = d["id"]
            if id not in self._devices:
                # try to match device name and combine with newer device
                for did in self._devices:
                    if self._devices[did].name == d["name"]:
                        id = did
                        break
                else:
                    self._devices[id] = SenseDevice(id)
                    self._devices[id].icon = d["icon"]
            if not self._devices[id].name:
                self._devices[id].name = d["name"]
            self._devices[id].energy_kwh[scale] += d["total_kwh"]

    @property
    def devices(self) -> list[SenseDevice]:
        """List of discovered device names."""
        return self._devices.values()

    def _set_realtime(self, data):
        """Sets the realtime data structure."""
        json_devices = data.get("devices", {})
        if not json_devices:
            return
        self._realtime = data
        for dev in self._devices.values():
            dev.is_on = False
            dev.power_w = 0
        for d in json_devices:
            id = d["id"]
            if id not in self._devices:
                self._devices[id] = SenseDevice(id)
            self._devices[id].power_w = float(d["w"])
            self._devices[id].is_on = self._devices[id].power_w > 0

    def get_realtime(self):
        """Outdated. Return the raw realtime data structure.
        Access sense.devices instead."""
        return self._realtime

    @property
    def active_power(self) -> float:
        return self._realtime.get("w", 0)

    @property
    def active_solar_power(self) -> float:
        return self._realtime.get("solar_w", 0)

    @property
    def active_voltage(self) -> float:
        return self._realtime.get("voltage", [])

    @property
    def active_frequency(self) -> float:
        return self._realtime.get("hz", 0)

    @property
    def daily_usage(self) -> float:
        return self.get_stat(Scale.DAY, "consumption")

    @property
    def daily_production(self) -> float:
        return self.get_stat(Scale.DAY, "production")

    @property
    def daily_production_pct(self) -> float:
        return self.get_stat(Scale.DAY, "production_pct")

    @property
    def daily_net_production(self) -> float:
        return self.get_stat(Scale.DAY, "net_production")

    @property
    def daily_from_grid(self) -> float:
        return self.get_stat(Scale.DAY, "from_grid")

    @property
    def daily_to_grid(self) -> float:
        return self.get_stat(Scale.DAY, "to_grid")

    @property
    def daily_solar_powered(self) -> float:
        return self.get_stat(Scale.DAY, "solar_powered")

    @property
    def weekly_usage(self) -> float:
        return self.get_stat(Scale.WEEK, "consumption")

    @property
    def weekly_production(self) -> float:
        return self.get_stat(Scale.WEEK, "production")

    @property
    def weekly_production_pct(self) -> float:
        return self.get_stat(Scale.WEEK, "production_pct")

    @property
    def weekly_net_production(self) -> float:
        return self.get_stat(Scale.WEEK, "net_production")

    @property
    def weekly_from_grid(self) -> float:
        return self.get_stat(Scale.WEEK, "from_grid")

    @property
    def weekly_to_grid(self) -> float:
        return self.get_stat(Scale.WEEK, "to_grid")

    @property
    def weekly_solar_powered(self) -> float:
        return self.get_stat(Scale.WEEK, "solar_powered")

    @property
    def monthly_usage(self) -> float:
        return self.get_stat(Scale.MONTH, "consumption")

    @property
    def monthly_production(self) -> float:
        return self.get_stat(Scale.MONTH, "production")

    @property
    def monthly_production_pct(self) -> float:
        return self.get_stat(Scale.MONTH, "production_pct")

    @property
    def monthly_net_production(self) -> float:
        return self.get_stat(Scale.MONTH, "net_production")

    @property
    def monthly_from_grid(self) -> float:
        return self.get_stat(Scale.MONTH, "from_grid")

    @property
    def monthly_to_grid(self) -> float:
        return self.get_stat(Scale.MONTH, "to_grid")

    @property
    def monthly_solar_powered(self) -> float:
        return self.get_stat(Scale.MONTH, "solar_powered")

    @property
    def yearly_usage(self) -> float:
        return self.get_stat(Scale.YEAR, "consumption")

    @property
    def yearly_production(self) -> float:
        return self.get_stat(Scale.YEAR, "production")

    @property
    def yearly_production_pct(self) -> float:
        return self.get_stat(Scale.YEAR, "production_pct")

    @property
    def yearly_net_production(self) -> float:
        return self.get_stat(Scale.YEAR, "net_production")

    @property
    def yearly_from_grid(self) -> float:
        return self.get_stat(Scale.YEAR, "from_grid")

    @property
    def yearly_to_grid(self) -> float:
        return self.get_stat(Scale.YEAR, "to_grid")

    @property
    def yearly_solar_powered(self) -> float:
        return self.get_stat(Scale.YEAR, "solar_powered")

    @property
    def active_devices(self):
        return [d.name for d in self._devices.values() if d.is_on]

    @property
    def time_zone(self) -> str:
        return self._monitor.get("time_zone", "")

    def trend_start(self, scale: Scale) -> Optional[datetime]:
        """Return start of trend last updated."""
        if "start" not in self._trend_data[scale]:
            return None
        try:
            return ciso8601.parse_datetime(self._trend_data[scale]["start"])
        except ValueError:
            pass
        return None

    def get_stat(self, scale: Scale, key: str) -> float:
        if scale not in self._trend_data or key not in self._trend_data[scale]:
            return 0
        data = self._trend_data[scale][key]
        if not isinstance(data, (dict, float, int)):
            return 0
        if isinstance(data, dict):
            return data.get("total", 0)
        return data

    def get_trend(self, scale: str, key: any) -> float:
        """Return trend data item from last update."""
        if isinstance(key, bool):
            key = "production" if key is True else "consumption"
        else:
            key = "consumption" if key == "usage" else key
        return self.get_stat(Scale[scale], key)



from setuptools import setup

with open("README.md", "r") as fh:
    long_description = fh.read()

version = "0.13.1"
setup(
    name="sense_energy",
    packages=["sense_energy"],
    install_requires=[
        "async_timeout>=3.0.1",
        "ciso8601",
        "orjson",
        "requests",
        "websocket-client",
        "websockets",
        "aiohttp",
        "kasa-crypt>=0.2.0",
    ],
    version=version,
    description="API for the Sense Energy Monitor",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="scottbonline",
    author_email="scottbonline@gmail.com",
    url="https://github.com/scottbonline/sense",
    download_url="https://github.com/scottbonline/sense/archive/" + version + ".tar.gz",
    keywords=["sense", "energy", "api"],
    classifiers=[
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "License :: OSI Approved :: MIT License",
    ],
)
